/*****************************************************************************
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: ir.h,v 1.19 2000/06/07 00:59:38 kevin Exp $
**  $Log: ir.h,v $
**  Revision 1.19  2000/06/07 00:59:38  kevin
**  added imagefiles
**
**  Revision 1.18  2000/06/05 01:32:45  kevin
**  Added C++ compatibility
**
**  Revision 1.17  2000/06/03 06:29:08  kevin
**  *** empty log message ***
**
**  Revision 1.16  2000/05/24 22:48:17  kevin
**  First functional version of SDF library for X-window
**
**  Revision 1.15  2000/05/16 04:33:17  kevin
**  Updated documentation
**
**  Revision 1.14  2000/05/11 14:07:00  kevin
**  Added support for Windows NT
**
**  Revision 1.13  2000/05/08 20:00:48  kevin
**  ANSI C changes
**
**  Revision 1.12  2000/05/07 12:46:19  kevin
**  made c++ compatible
**
**  Revision 1.11  2000/05/05 02:37:31  kevin
**  renamed phmelm to pelm
**
**  Revision 1.10  2000/05/04 18:16:34  kevin
**  renamed filter definitions
**
**  Revision 1.9  2000/05/04 04:29:18  kevin
**  *** empty log message ***
**
**  Revision 1.8  2000/05/04 04:25:55  kevin
**  Renamed phantom and phantom-element functions/variables
**
**  Revision 1.7  2000/05/03 19:51:41  kevin
**  function renaming for phantoms and phantom elements
**
**  Revision 1.6  2000/05/03 08:49:49  kevin
**  Code cleanup
**
**  Revision 1.5  2000/05/02 20:00:25  kevin
**  *** empty log message ***
**
**  Revision 1.4  2000/05/02 15:31:39  kevin
**  code cleaning
**
**  Revision 1.3  2000/04/29 23:24:29  kevin
**  *** empty log message ***
**
**  Revision 1.2  2000/04/28 14:14:16  kevin
**  *** empty log message ***
**
**  Revision 1.1.1.1  2000/04/28 13:02:43  kevin
**  Initial CVS import for first public release
**
**
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

/*	ir.h	    Header File for Image Reconstruction System
 *	Programmer:   Kevin Rosenberg
 *	Date Started: 7-1-84
 */

#ifndef IR_H
#define IR_H

#ifdef MPI_CT
#define MPI_MAX_PROCESS 128 
struct mpi_ct_st
{
  int my_rank;
  int nproc;
  int base_local_work_units;
  int remainder_work_units;
  int local_work_units[MPI_MAX_PROCESS];
  int start_work_unit[MPI_MAX_PROCESS];
  MPI_Comm comm;
};

extern struct mpi_ct_st mpi_ct;
void mpi_ct_calc_work_units(const unsigned int global_work_units);
#endif


struct histo_st {
    int *b;			/* Histogram array (# of elements in each bin) */
    int nbin;			/* Number of histogram bins */
    double xmin, xmax, xinc;	/* Limits of histogram boundaries */
};

typedef struct histo_st HISTOGRAM;

/*---------------------------------------------------------------------------*/

static const int POINTS_PER_CIRCLE=90;
#define MAXREMARK  99

typedef FMTX_2D         IMAGE_ARRAY;    /* use 2d floating point matrix */
typedef float           IMAGE_ELEM_VAL;	/* use floats for image storing */

struct image_st {
    IMAGE_ARRAY v;			/* values of voxels in matrix form */
    SDF_2D *dfp_2d;			/* Pointer to disk image file */
    int nx, ny;				/* size of voxel matrix */
    double xmin, xmax, ymin, ymax;	/* extent of voxel matrix in phm coord */
    char remark[MAXREMARK+1];		/* description of voxel data */
    float calctime;			/* time to calculate voxels in seconds */
};
typedef struct image_st  IMAGE;

typedef enum {
  RECTANGLE,
  TRIANGLE,
  ELLIPSE,
  SECTOR,
  SEGMENT
} PElmType;

struct pelm_st {
    PElmType type;	     /* pelm type (box, ellipse, etc) */
    double atten;	     /* X-ray attenuation coefficient */
    double cx,cy;	     /* center of pelm */
    double u,v;		     /* size of pelm */
    double rot;		     /* pelm rotation angle (in radians) */
    double *x, *y;	     /* ptr to array of points in obj world coord */
    int pts;		     /* number of points in outline arrays */
    double xmin, xmax, ymin, ymax; /* pelm limits */
    double radius;	     /*   "   */
    GRFMTX_2D p_to_o;        /* map from phantom to standard pelm coords */
    GRFMTX_2D o_to_p;        /* map from std pelm coords to phantom coords */
    struct pelm_st *next;    /* pointer to next pelm in phantom */
};
typedef struct pelm_st PELM;

typedef enum {
  P_PELMS,        /* Phantom made of Pelms */
  P_UNIT_PULSE,   /* Special phantom, not made of pelms */
  P_FILTER	  /* defined only by this type */
} PhmType;

struct phm_st {		           /* Phantom structure */
    PELM *pelm_list;               /* pelm linked-list */
    PhmType type;
    int n_pelm;		           /* number of pelms in phantom */
    double xmin, xmax, ymin, ymax; /* extent of pelms in pelm coordinates */
    double radius;                 /*      " "    */
};
typedef struct phm_st    PHANTOM;


/*----------------------------------------------------------------------*/
/*				RAYSUM SYMBOLS				*/
/*----------------------------------------------------------------------*/

/* Ray sums are collected along an array of ndet detectors.  The data
 * for these detectors is stored in the structure DETECTARRAY
 */

typedef float DETECT_TYPE;

struct detarray_st {
  DETECT_TYPE *detval;	/* Pointer to array of values recorded by detector */
  int ndet;		/* Number of detectors in array */
  double view_angle;    /* View angle in radians */
};
typedef struct detarray_st DETARRAY;


typedef enum {
    DETECTOR_PARALLEL,
    DETECTOR_EQUIANGLE,
    DETECTOR_EQUILINEAR
} ScannerGeometry;
  
struct detector_st {
  ScannerGeometry geometry;     /* Geometry of detectory */
  int ndet;			/* Number of detectors in array */
  int nview;			/* Number of rotated views */
  int nsample;		        /* Number of rays per detector */
  double detlen;		/* Total length of detector array */
  double rotlen;		/* Rotation angle length in radians (norm 2PI) */
  double det_inc;		/* Increment between centers of detectors */
  double rot_inc;		/* Increment in rotation angle between views */
  double radius;		/* Radius of rotation.  Distance from */
				/*   center of phm to center of det */
  double phmlen;                /* Maximum Length of phantom or area of interest */
  struct {
    double xd1,yd1,xd2,yd2;	/* Coordinates of detector endpoints */
    double xs1,ys1,xs2,ys2;	/* Coordinates of source endpoints */
    double angle;		/* Starting angle */
  } init;
};
typedef struct detector_st DETECTOR;

struct raysum_st {
  int fd;			/* Disk file descriptor */
  int file_mode;		/* Current file mode (read or write) */
  int header_size;		/* Size of disk file header */
  int geometry;			/* Geometry of scanner */
  struct detarray_st **view;	/* Pointer to array of detarray_st pointers */

  char remark[MAXREMARK+1];	/* description of raysum data */
  double calctime;		/* time required to calculate raysums */

  int ndet;			/* number of detectors in array */
  int nview;			/* number of rotated views */
  double rot_start;		/* starting view rotation */
  double rot_inc;		/* angle between rotations */
  double det_start;		/* distance of beginning detector to center */
				/*    of PHANTOM */
  double det_inc;		/* increment between detectors */
  double phmlen;		/* Length of PHANTOM edge (phm is square) */
};
typedef struct raysum_st   RAYSUM;

/*----------------------------------------------------------------------*/
/*				USER SYMBOLS				*/
/*----------------------------------------------------------------------*/

/* Codes for Coordinate Types      */
/* Defines coords for pelm_is_point_inside() */

typedef enum {
  PELM_COORD,         /* Normalized Pelm Coordinates */
  PHM_COORD           /* User's phantom Coordinates */
} CoordType;

/* Codes for Filter types */

typedef enum { 	/* filter types for filter_generate() */
  FILTER_BANDLIMIT, 
  FILTER_SINC,
  FILTER_G_HAMMING,
  FILTER_COSINE,
  FILTER_TRIANGLE,
  FILTER_ABS_BANDLIMIT, 	/* filter times |x| */
  FILTER_ABS_SINC, 
  FILTER_ABS_G_HAMMING,
  FILTER_ABS_COSINE,
  FILTER_SHEPP
} FilterType;

/* function domains */

static const char D_FREQ_STR[]=    "freq";
static const char D_SPATIAL_STR[]= "spatial";
 
typedef enum {
  D_FREQ = 1,
  D_SPATIAL 
} DomainType;

typedef enum {
    FUNC_EVEN = 1,    /* function types, f[-n] = f[n] */
    FUNC_ODD,     /* f[-n] = -f[n] */
    FUNC_BOTH    /* function has both odd & even components */
} FunctionSymmetry;

/* interpolation methods */
#undef HAVE_BSPLINE_INTERP
typedef enum {     /* Interpolation methods */
  I_NEAREST = 1,       /* Nearest neighbor */
#if HAVE_BSPLINE_INTERP
  I_BSPLINE,
  I_1BSPLINE,      /* 1st order B-Spline */
  I_2BSPLINE,
  I_3BSPLINE,
#endif
  I_LINEAR        /* Linear interpolation */
} InterpolationType;

/* Constants for sizing PHANTOM */

static const double PERCENT_PHM_SIZE_INCR=0.0;  /* Fractional increase in phantom limits compared to pelm size */
static const int N_EXTRA_DETECTORS=4;           /* Number of extra detectors widths when calculating detlen */

static const char O_TRACE_NONE_STR[]=     "none";
static const char O_TRACE_TEXT_STR[]=     "text";
static const char O_TRACE_PHM_STR[]=      "phm";
static const char O_TRACE_RAYS_STR[]=     "rays";
static const char O_TRACE_PLOT_STR[]=     "plot";
static const char O_TRACE_CLIPPING_STR[]= "clipping";

enum {
  TRACE_NONE, 		/* No tracing */
  TRACE_TEXT,		/* Minimal status */
  TRACE_PHM,		/* Show phantom */
  TRACE_RAYS,		/* Show all rays */
  TRACE_PLOT,		/* Plot raysums */
  TRACE_CLIPPING	/* Plot clipping */
};

typedef enum {
  O_PHM_HERMAN,               /* Herman head phantom */
  O_PHM_ROWLAND,              /* Rowland head phantom */
  O_PHM_BROWLAND,             /* Bordered Rowland head phantom */
  O_PHM_UNITPULSE             /* Unit pulse phantom */
} PhantomType;

static const char O_PHM_HERMAN_STR[]=    "herman";
static const char O_PHM_ROWLAND_STR[]=   "rowland";
static const char O_PHM_BROWLAND_STR[]=  "browland";
static const char O_PHM_UNITPULSE_STR[]= "unitpulse";

static const char O_INTERP_NEAREST_STR[]=  "nearest";
static const char O_INTERP_LINEAR_STR[]=   "linear";
static const char O_INTERP_BSPLINE_STR[]=  "bspline";

static const char O_FILTER_ABS_BANDLIMIT_STR[]= "abs_bandlimit";
static const char O_FILTER_ABS_SINC_STR[]=      "abs_sinc";
static const char O_FILTER_ABS_COS_STR[]=       "abs_cos";
static const char O_FILTER_ABS_HAMMING_STR[]=   "abs_hamming";
static const char O_FILTER_SHEPP_STR[]=         "shepp";
static const char O_FILTER_BANDLIMIT_STR[]=     "bandlimit";
static const char O_FILTER_SINC_STR[]=          "sinc";
static const char O_FILTER_COS_STR[]=           "cos";
static const char O_FILTER_HAMMING_STR[]=       "hamming";
static const char O_FILTER_TRIANGLE_STR[]=      "triangle";

typedef  enum {
  O_BPROJ_TRIG,
  O_BPROJ_TABLE,
  O_BPROJ_DIFF,
  O_BPROJ_DIFF2,
  O_BPROJ_IDIFF2
} BackprojType;

static const char O_BPROJ_TRIG_STR[]=     "trig";
static const char O_BPROJ_TABLE_STR[]=    "table";
static const char O_BPROJ_DIFF_STR[]=     "diff";
static const char O_BPROJ_DIFF2_STR[]=    "diff2";
static const char O_BPROJ_IDIFF2_STR[]=   "idiff2";

const static int RAYSUM_TRACE_ROW_TITLE=1;
const static int RAYSUM_TRACE_ROW_TITLE2=2;
const static int RAYSUM_TRACE_ROW_PHANT_ID=4;
const static int RAYSUM_TRACE_ROW_CHROMATIC=7;
const static int RAYSUM_TRACE_ROW_SCATTER=8;
const static int RAYSUM_TRACE_ROW_PHOT_STAT=9;
const static int RAYSUM_TRACE_ROW_NDET=12;
const static int RAYSUM_TRACE_ROW_NVIEW=13;
const static int RAYSUM_TRACE_ROW_SAMPLES=14;
const static int RAYSUM_TRACE_ROW_CURR_VIEW=17;
const static int RAYSUM_TRACE_ROW_ATTEN=18;


/*************************************************************************
 *  FUNCTION DECLARATIONS
 ************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

void usage (const char *program);

/* From reconstr.c */
IMAGE *image_reconst (IMAGE *im, RAYSUM *rs, const FilterType filt_type, double filt_param, InterpolationType interp_type, int interp_param, const BackprojType backproj_type, int const ir_trace);

/* From bproj.c */
void backproj_init (const RAYSUM *rs, IMAGE *im, const BackprojType bproj_method);
int  backproj_calc (const RAYSUM *rs, IMAGE *im, const double *t, const double view_angle, 
		    const int interp_type, const int bproj_method);
void backproj_term (const RAYSUM *rs, IMAGE *im, const int bproj_method);

void backproj_init_trig (const RAYSUM *rs, IMAGE *im);
int  backproj_calc_trig (const RAYSUM *rs, IMAGE *im, const double *t, 
			 const double view_angle, const int interp_type);
void backproj_term_trig (const RAYSUM *rs, IMAGE *im);
void backproj_init_table (const RAYSUM *rs, IMAGE *im);
int  backproj_calc_table (const RAYSUM *rs, IMAGE *im, const double *t, 
			  const double view_angle, const int interp_type);
void backproj_term_table (const RAYSUM *rs, IMAGE *im);
void backproj_init_d (const RAYSUM *rs, IMAGE *im);
int  backproj_calc_d (const RAYSUM *rs, IMAGE *im, const double *t, 
		      const double view_angle, const int interp_type);
void backproj_term_d (const RAYSUM *rs, IMAGE *im);
void backproj_init_d2 (const RAYSUM *rs, IMAGE *im);
int  backproj_calc_d2 (const RAYSUM *rs, IMAGE *im, const double *t, 
		       const double view_angle, const int interp_type);
void backproj_term_d2 (const RAYSUM *rs, IMAGE *im);
void backproj_init_id (const RAYSUM *rs, IMAGE *im);
int  backproj_calc_id (const RAYSUM *rs, IMAGE *im, const double *t, 
		       const double view_angle, const int interp_type);
void backproj_term_id (const RAYSUM *rs, IMAGE *im);
void backproj_init_id2 (const RAYSUM *rs, IMAGE *im);
int  backproj_calc_id2 (const RAYSUM *rs, IMAGE *im, const double *t, 
			const double view_angle, const int interp_type);
void backproj_term_id2 (const RAYSUM *rs, IMAGE *im);

/* bspline.c */
int bspline(int samples, int zoom_factor, int spline_order, double input[], double output[]);

/* convolve.c */
double convolve(const double f1[], const double f2[], const double dx, const int n, const int np, const FunctionSymmetry func_type);
double convolve_both(const double f1[], const double f2[], const double dx, const int n, const int np);

/* dialogs.c */
int phm_add_pelm_kb(PHANTOM *phm);
PHANTOM *phm_select(void);
int interpolation_select(void);
int filter_select(double *filter_param);

/* filter.c */
double *filter_generate(const FilterType filt_type, double bw, double xmin, double xmax, int n, double param, const DomainType domain, int numint);
double filter_spatial_response_calc(int filt_type, double x, double bw, double param, int n);
double filter_spatial_response_analytic(int filt_type, double x, double bw, double param);
double filter_frequency_response(int filt_type, double u, double bw, double param);
double sinc(double x, double mult);
double integral_abscos(double u, double w);

/* image.c */
IMAGE *image_create(const char *fname, const int nx, const int ny);
int image_clear(IMAGE *im);
int image_save(IMAGE *im);
IMAGE *image_load(const char *fname);
void image_filter_response(IMAGE *im, const DomainType domain, double bw, const FilterType filt_type, double filt_param, const int opt_trace);
int image_display (const IMAGE *im);
int image_display_scale (const IMAGE *im, const int scale, const double pmin, const double pmax);

/* options.c */
int opt_set_trace(const char *optarg);
const char *name_of_phantom(const int phmid);
int opt_set_phantom(const char *optarg);
int opt_set_interpolation(const char *optarg);
const char *name_of_interpolation(int interp_type);
int opt_set_filter(const char *optarg);
const char *name_of_filter(const int filter);
DomainType opt_set_filter_domain(const char *optarg);
const char *name_of_filter_domain(const DomainType domain);
int opt_set_backproj(const char *optarg);
const char *name_of_backproj(const BackprojType backproj);

/* phm.c */
PHANTOM *phm_create(const int phmid);
PHANTOM *phm_create_from_file(const char *fname);
PHANTOM *phm_init(void);
void phm_free (PHANTOM *phm);
int phm_add_pelm_file(PHANTOM *phm, const char *fname);
void phm_add_pelm (PHANTOM *phm, const char *type, const double cx, const double cy, 
	       const double u, const double v, const double rot, const double atten);
int pelm_make_points(PELM *obj);
void pelm_make_xform (PELM *obj);
PELM *pelm_alloc(void);
void calc_arc(double x[], double y[], const int pts, const double xcent, const double ycent, 
	      const double r, const double start, const double stop);
void calc_ellipse(double x[], double y[], const int pts, const double u, const double v);
int circle_pts(double theta);
void phm_print(PHANTOM *phm);
#if HAVE_SGP
void phm_show(const PHANTOM *phm);
void phm_draw(const PHANTOM *phm);
#endif

/* phm2image.c */
void phm_to_image(const PHANTOM *phm, IMAGE *im, const int col_start, const int col_count,
		  const int nsample, const int trace);
int pelm_is_point_inside(PELM *obj, const double x, const double y, const CoordType coord_type);

/* phmstd.c */
void phm_std_herman (PHANTOM *phm);
void phm_std_rowland (PHANTOM *phm);
void phm_std_rowland_bordered (PHANTOM *phm);

/* raycollect.c */
int raysum_collect(RAYSUM *rs, const DETECTOR *det, const PHANTOM *phm, const int start_view, const int trace, const int unit_pulse);
void rayview(const PHANTOM *phm, DETARRAY *darray, const DETECTOR *det, const double xd1, const double yd1, const double xd2, const double yd2, const double xs1, const double ys1, const double xs2, const double ys2, const int unit_pulse);
double phm_ray_attenuation (const PHANTOM *phm, const double x1, const double y1, const double x2, const double y2);
double pelm_ray_attenuation (PELM *pelm, const double x1, const double y1, const double x2, const double y2);
int pelm_clip_line (const PELM *pelm, double *x1, double *y1, double *x2, double *y2);
void raysum_trace_show_param (const char *label, const char *fmt, int row, int color, ...);

/* scanner.c */
DETECTOR *detector_create(const PHANTOM *phm, int geometry, int ndet, int nview, int nsample, const double rot_anglen);
void detector_free(DETECTOR *det);

/* rayio.c */
RAYSUM *raysum_create(const char *fname, const int nview, const int ndet);
RAYSUM *raysum_create_from_det(const char *fname, const DETECTOR *det);
RAYSUM *raysum_open(const char *filename);
void raysum_alloc_views(RAYSUM *rs);
void raysum_free(RAYSUM *rs);
int raysum_is_open(RAYSUM *rs);
int raysum_close(RAYSUM *rs);
int raysum_read_header(RAYSUM *rs);
int raysum_write_header(RAYSUM *rs);
int raysum_read(RAYSUM *rs);
int raysum_write(RAYSUM *rs);
DETARRAY *detarray_alloc(const int n);
void detarray_free(DETARRAY *darray);
int detarray_read(RAYSUM *rs, DETARRAY *darray, const int view_num);
int detarray_write(RAYSUM *rs, const DETARRAY *darray, const int view_num);
int raysum_print(const RAYSUM *rs);

#ifdef __cplusplus
} 
#endif /* __cplusplus */

#endif
